/*
       The Travelling Salesman Problem Solver

       Path Improvement

       Given a path, this module will attempt to improve that path.

       On the Hypercube, one copy of this module executes in each
       node. 
*/	
#include <stdio.h>		       /* ***TEMP*** not on Hypercube */
#include <assert.h>
#include "config.h"
#include "travel.h"
extern boolean debug;		       /* debugging */
extern boolean verbose;		       /* show improvements */
extern long distance();		       /* external */
static struct path *pathp;	       /* visible paths	*/
extern long lsqrt();		       /* long integer square root */
struct best {			       /* best-path information	*/
       long bval;		       /* best delta distance */
       short brev;		       /* reversal flags */
       };
/*
       logchg  --  log a change

       ***TEMP***
*/
static void logchg(bk1,bk2,inter,revpat,olddist,newdist)       
short bk1,bk2;		       /* breakpoints */
short inter;		       /* which	permutation */
short revpat;		       /* reversal pattern */
long olddist,newdist;	       /* distances before and after */
{
       printf("Improvement: %ld=>%ld",
	       olddist,newdist);
       if (!debug)	       /* if not debugging */
       {       printf("\n");   /* finish line */
	       return;
       }
       printf(" [0..%d] [%d..%d] [%d..%d]  ",
	       bk1,bk1+1,bk2-1,bk2,getcnt()-1);
       if (revpat & 4) printf("A");    /* edit reversal pattern */
       if (revpat & 2) printf("B");
       if (revpat & 1) printf("C");
       printf("  ");
       switch (inter) {		       /* edit permutation pattern */
       case 0: break; 
       case 1: printf("B:C"); break;
       case 2: printf("A:B"); break;
       };  

       printf("\n");
}
/*
       logstats	 --  log statistics

       ***TEMP*** must send message on Hypercube
*/
static void logstats(tries,changes,maxsince)
long tries;			       /* total	tries */
long changes;			       /* total	changes	*/
short maxsince;			       /* max since last change	*/
{
       printf("%ld tries, %ld changes, %d longest run without improvement.\n",
		       tries,changes,maxsince);	/* ***TEMP*** */
}
/*
       doreverse  --  reverse the indicated portion of the path
*/
static void doreverse(pathp,bounds)
struct path *pathp;		       /* path to work on */
short bounds[2];		       /* bounds of reversal area */
{      short i,j;		       /* for loop */
       short temp;		       /* for swap */
       short lo	= bounds[0];	       /* low bound */
       short hi	= bounds[1];	       /* high bound */
       assert(lo <= hi);	       /* bounds valid */
       assert(lo >= 0);
       assert(hi < pathp->pathn);      /* check	against	length */
       for (i=lo, j=hi;	i<j; i++,j--)  /* do swap */
       {       temp = pathp->plinks[i];/* get one */
	       pathp->plinks[i]	= pathp->plinks[j];
	       pathp->plinks[j]	= temp;
       }
}
/*
       dointer	--  interchange	two areas

       The areas need not be the same length, but they are
       always adjacent.	 We can	thus accomplish	the move by
       reversing both sections,	then reversing the result.

       Since we	may have reversed sections above, this is    
       inefficient and we can speed it up with some bookeeping above. 
*/
static void dointer(pathp,b1,b2)
struct path *pathp;		       /* path to work on */
short b1[2],b2[2];		       /* bounds of areas to interchange */
{      short bwork[2];		       /* bounds of both */
       assert(b1[1]+1 == b2[0]);
       doreverse(pathp,b1);	       /* reverse part 1 */
       doreverse(pathp,b2);	       /* reverse part 2 */
       bwork[0]	= b1[0];	       /* construct full bounds	*/
       bwork[1]	= b2[1];
       doreverse(pathp,bwork);	       /* reverse entire */
}
/*
       tryperm	--  try	a permutation

       There are eight reversals to try	for each permutation.
       Note that the index of the reversal chosen is also a
       bit-coded representation	of the reversals required.
*/
static void tryperm(patha,pathb,pathc,bestp)
short patha[2];				       /* endpoint cities of path A */
short pathb[2];				       /* endpoint cities of path B */
short pathc[2];				       /* endpoint cities of path C */
struct best *bestp;			       /* best info OUT	*/
{
#define	REV(p) { temp =	p[0]; p[0] = p[1]; p[1]	= temp;	} /* reverse */
       short temp;			       /* for reversal macro */
       long bestval = 0x7ffffffL;	       /* best value found so far */
       short i,j,k;			       /* for loops */
       long dist;			       /* working distance */
       short bestr = -1;		       /* best reversal	*/
       short n=0;			       /* reversal index */
       for (i=0; i<2; i++)		       /* try reversals	of A */
       {       for (j=0; j<2; j++)	       /* try reversals	of B */
	       {       for (k=0; k<2; k++)     /* try reversals	of C */
		       {       /*      Compute distance	for current rev	*/
			       dist = distance(patha[1],pathb[0]) +
				      distance(pathb[1],pathc[0]);
			       if (dist	< bestval) /* if new best, save	*/
			       {       bestval = dist;
				       bestr = n;
			       }
			       n++;
			       REV(pathc);     /* reverse C */
		       }
		       REV(pathb);	       /* reverse B */
	       }
	       REV(patha);		       /* reverse A */
       }
       assert(n	== 8);			       /* tried	all? */
       assert(bestval >= 0);		       /* got something? */
       assert(bestr >= 0);		       /* saved	it? */
       bestp->bval = bestval;		       /* return best value */
       bestp->brev = bestr;		       /* return best rev pattern */
}
/*
       iiter  --  one iteration	of improve

       The algorithm is	as follows:

       1.  Pick	two links at random. 
       2.  Consider the	three subpaths that remain with	the selected
	   links broken.  Compute the total path length	for all	possible
	   rearrangements of the given subpaths.  Save the identity of
	   the shortest	rearrangement.
       3.  If the shortest rearrangement is shorter than the length of
	   the original	path, update the path accordingly and return TRUE.
       4.  Otherwise do	not alter the path and return FALSE.

       There are 24 possible permutations, constructed as follows:
       There are three subpaths, which can be arranged in the ways
	       ABC, ACB, BAC.
       For each	arrangement, each path can be reversed,	so there are
       8 possible reversal arrangements	for each path.
*/
static boolean iiter()			       /* returns TRUE if improved */
{
       short cnt = pathp->pathn;	       /* number of cities */
       long olddist = pathp->pathl;	       /* length of path */
       long newdist;			       /* new length of	path */
       long oldval;			       /* old dist across cuts */
       long bestval;			       /* best dist across cuts	*/
       short bk1,bk2;			       /* indices of breakpoints */
       short spa[2],spb[2],spc[2];	       /* end cities of	subpaths */
       short bounda[2],boundb[2],boundc[2];    /* path indices of subpaths */
       struct best best1,best2,best3;	       /* best info */
       short revpat;			       /* reversal pattern */
       short inter;			       /* interchange plan */
       /*      Divide the path into three subpaths, each with at least */
       /*      two cities.					       */
       bk1 = irandom(0,cnt-3);		       /* in [0..cnt-3]	*/
       bk2 = irandom(bk1+2,cnt-1);	       /* in [bk1+1..cnt-1] */
       if (debug)
               printf("Breaks at %d and %d\n",bk1,bk2);
       assert(bk1 >= 0);
       assert(bk2 > bk1);
       assert(cnt-1 >= bk2);
       /*      Get endpoints of	sections.  These are path indices.  */
       bounda[0] = 0;			       /* subpath A */
       bounda[1] = bk1;
       boundb[0] = bk1+1;		       /* subpath B */
       boundb[1] = bk2-1;			
       boundc[0] = bk2;			       /* subpath C */
       boundc[1] = cnt-1;
       /*      Get endpoints of	sections.  These are city indices.  */
       spa[0] =	pathp->plinks[0];	       /* first	of subpath A, first */
       spa[1] =	pathp->plinks[bk1];	       /* last of subpath A */
       spb[0] =	pathp->plinks[bk1+1];	       /* first	of subpath B */
       spb[1] =	pathp->plinks[bk2-1];	       /* last of subpath B */
       spc[0] =	pathp->plinks[bk2];	       /* first	of subpath C */
       spc[1] =	pathp->plinks[cnt-1];	       /* last of subpath C, last */
       oldval =	distance(spa[1],spb[0])	+      /* unpermuted distance value */
		distance(spb[1],spc[0]);
       /*      Try all 8 reversal patterns for each permutation.       */
       tryperm(spa,spb,spc,&best1);	       /* unpermuted (ABC) */
       tryperm(spa,spc,spb,&best2);	       /* reverse B and	C */
       tryperm(spb,spa,spc,&best3);	       /* reverse B and	A */
       bestval = min(best1.bval,min(best2.bval,best3.bval));
       if (bestval == oldval) return(FALSE);   /* no improvement */
       assert(bestval <	oldval);	       /* should not fail, we try original */
       /*      The situation has improved.  We will rearrange the path.	*/  
       revpat =	0;			       /* construct reversal pattern */
       if (bestval == best1.bval)
       {
	       inter = 0;		       /* interchange nothing */
	       revpat =	best1.brev;	       /* save reversal	pattern	*/
	       assert(revpat);		       /* if no	swap, must reverse */
       } else if (bestval == best2.bval)
       {       inter = 1;		       /* interchange B	and C */
	       if (best2.brev &	4) revpat |= 4;/* in reversal pattern too */
	       if (best2.brev &	2) revpat |= 1;
	       if (best2.brev &	1) revpat |= 2;
       } else {assert(bestval == best3.bval);  /* must be option 3 */
	       inter = 2;		       /* interchange A	and B */
	       if (best3.brev &	4) revpat |= 2;/* in reversal pattern too */
	       if (best3.brev &	2) revpat |= 4;
	       if (best3.brev &	1) revpat |= 1;
       }
       /*      Implement the planned path improvement. */
       newdist = olddist - oldval + bestval;   /* new distance */
       logchg(bk1,bk2,inter,revpat,olddist,newdist);   /* log change */
       if (revpat & 1) doreverse(pathp,boundc);	  /* implement reversals */
       if (revpat & 2) doreverse(pathp,boundb);	  /* implement reversals */
       if (revpat & 4) doreverse(pathp,bounda);	  /* implement reversals */
       switch (inter) {			       /* implement interchanges */
       case 0: break;			       /* nothing */
       case 1: dointer(pathp,boundb,boundc);  break; /*	interchange B and C */
       case 2: dointer(pathp,bounda,boundb);  break; /*	interchange A and B */
       };
       pathp->pathl = newdist;		       /* update path length */
       return(TRUE);			       /* path rearranged */
}
/*
       improve	--  improve a path

       Runs until an improvement has been detected or until the
       specified number	of iterations have been	completed without
       improvement.
*/
void improve(pp)
struct path *pp;			       /* Path info IN/OUT */
{
       long tries = 0;			       /* tries	made */
       long changes = 0;		       /* changes made */
       short sincechg =	0;		       /* tries	since last change */
       short maxsince =	0;		       /* max of sincechg at change */
       pathp = pp;			       /* for subroutines here */
       for (;;)			       /* until	limit or chg */
       {       tries++;		       /* tally	tries */
	       if (iiter())	       /* try the algorithm */
	       {
		       if (verbose)
		       {       printpath(stdout,pathp);	/* print it */
		       }
		       if (debug)
		       {
			       checkpath(pathp); /* check result */
		       }
		       maxsince	= max(maxsince,sincechg);
		       sincechg	= 0;   /* no tries since chg */
		       changes++;      /* tally	changes	*/
		       logstats(tries,changes,maxsince);
		       continue;       /* keep trying */
	       }
	       sincechg++;	       /* tally	since change */
	       /*      Termination conditions:	       */
	       /*      Always try at least cities*3.   */
	       /*      No improvement for more than    */
	       /*      3x the previous longest run of  */
	       /*      fails ends the computation.     */
	       if (sincechg < pathp->pathn*3) continue;	
	       if (sincechg > 3*maxsince) break;
       }
       logstats(tries,changes,maxsince);       /* final	statistics */
       return;				       /* done */
}
